/*
   Copyright (C) 1996 Csar Crusius
   Modified      1997 Helmut Bohr to use with the FORMS LIBRARY  

   This file is part of the DND Library.  This library is free
   software; you can redistribute it and/or modify it under the terms of
   the GNU Library General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your
   option) any later version.  This library is distributed in the hope
   that it will be useful, but WITHOUT ANY WARRANTY; without even the
   implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU Library General Public License for more details.
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* TAB SIZE = 8 */


#include "FlDnD.h"
#include <X11/cursorfont.h>
#include <X11/Xatom.h>
#include <X11/Xmu/WinUtil.h>
#include <stdio.h>
#include <stdlib.h>
#include <values.h>
#include <forms.h>

/* Local variables */
static Display		*dpy;		/* current display		*/
static XButtonEvent	StartEvent;	/* event that started the drag	*/
static int		DragPrecision;	/* minimum dx,dy to start drag	*/
static int		Dragging;	/* Drag state flag		*/
static int		DataOK;		/* Non-zero if data registered	*/
static Atom		DndProtocol;	/* ClientMessage identifier	*/
static Atom		DndSelection;	/* For the data transfers	*/
static Atom		WM_STATE;	/* Needed for icon stuff	*/
static Window		Target;		/* Drop window			*/
static int		DataType;	/* Current drag data type	*/
static int		RootFlag;	/* Non-zero if dropped on root	*/

/*
   * Data for the Dnd cursors
*/
#include "cursor/MASK.xbm"
#include "cursor/file.xbm"
#include "cursor/file_mask.xbm"
#include "cursor/files.xbm"
#include "cursor/files_mask.xbm"
#include "cursor/dir.xbm"
#include "cursor/dir_mask.xbm"
#include "cursor/text.xbm"
#include "cursor/text_mask.xbm"
#include "cursor/grey.xbm"
#include "cursor/link.xbm"
#include "cursor/app.xbm"

typedef struct
{     int	Width,Height;
      char	*ImageData,*MaskData;
      int	HotSpotX,HotSpotY;
      Pixmap	ImagePixmap,MaskPixmap;
      Cursor	CursorID;
} CursorData;

static CursorData DndCursor[DndEND]={
				       {  0,0,NULL,NULL,0,0,0  },
				       {  grey_width,	grey_height,	grey_bits,	MASK_bits	,16,16  },
				       {  file_width, file_height,	file_bits,	file_mask_bits	,16,16  },
				       {  files_width,files_height,	files_bits,	files_mask_bits	,16,16  },
				       {  text_width,text_height,text_bits,text_mask_bits,text_x_hot,text_y_hot  },
				       {  dir_width,	dir_height,     dir_bits,       dir_mask_bits	,16,16  },
				       {  link_width,	link_height,	link_bits,	MASK_bits	,16,16  },
				       {  app_width,  app_height,     app_bits,       MASK_bits	,16,16  }
				    };

/* FlDndStartAction
   * 
   * Initializes the drag and drop structures. Binded to ButtonPress event in
   * registered drag widgets.
   * 
   * Last modification
   * 
   * Feb 01 1996 (v0.0) : created
   * Apr 01 1997 (v0.0) : modified to use with FORMS LIBRARY
*/
void   FlDndStartAction(FL_OBJECT *ob, XEvent *event)
   {
      StartEvent=*((XButtonEvent *)event);
      Dragging=0; DataOK=0;
      
#ifdef VERBOSE
      fprintf(stderr,"Initializing drag and drop variables...\n");
      fprintf(stderr,"Root 0x%x Window 0x%x Subwindow 0x%x\n",
	 StartEvent.root,StartEvent.window,StartEvent.subwindow);
#endif
   }

/* FlDndHandleDragging
   *
   * Takes care of the drag and drop process. Wait until the pointer had moved
   * a little. Then takes control over the pointer until the buttons are
   * released. After that send a Drag And Drop ClientMessage event. Returns
   * non-zero if a drop did take place.
   *
   * Last modification
   *
   * Feb 01 1996 (v0.0) : created
   * Apr 01 1997 (v0.0) : modified to use with FORMS LIBRARY
*/
int   FlDndHandleDragging(FL_OBJECT *ob,XEvent *event)
   {
      XEvent	Event;
      Window	root	= DefaultRootWindow(dpy);
      Window  DispatchWindow;

      if(Dragging) return 0;
      if(DataOK==0) return 0;
      if(abs(StartEvent.x_root - event->xmotion.x_root) < DragPrecision &&
	 abs(StartEvent.y_root - event->xmotion.y_root) < DragPrecision)
	 return 0;

#ifdef VERBOSE
      fprintf(stderr,"Initializing drag...\n");
#endif
      XUngrabPointer(dpy,CurrentTime);
      XGrabPointer(dpy,root,False,
	 ButtonMotionMask|ButtonPressMask|ButtonReleaseMask,
	 GrabModeSync,GrabModeAsync,root,
	 DndCursor[DataType].CursorID,
	 CurrentTime);

      Dragging=1; RootFlag=0;

      while (Dragging)
      {
	 XAllowEvents(dpy,SyncPointer,CurrentTime);
	 XNextEvent(dpy,&Event);
#ifdef VERBOSE
	 printf("Dragging :%d\n",Event.type);
#endif
	 switch (Event.type)
	 {
	    case ButtonRelease :
               if (Event.xbutton.subwindow) RootFlag = 0;
               else 	                    RootFlag = 1;
	       Dragging = 0;
#ifdef VERBOSE
               printf("DraggingE:%d\n",Event.type);
#endif
	       break;
	 }
      }
      XUngrabPointer(dpy,CurrentTime);
      
#ifdef VERBOSE
	 printf("RootFlag :%d\n",RootFlag);
#endif

      if(!RootFlag)
      {
	 Target=XmuClientWindow(dpy,Event.xbutton.subwindow);
#ifdef VERBOSE
	 if(Target==Event.xbutton.subwindow)
	    fprintf(stderr,"Dropped on an icon.\n");
#endif
	 if(Target==Event.xbutton.subwindow)
	    DispatchWindow=Target;
	 else	DispatchWindow=PointerWindow;
      }
      else
	 Target=DispatchWindow=ob->form->window;

      Event.xclient.type		= ClientMessage;
      Event.xclient.display		= dpy;
      Event.xclient.message_type	= DndProtocol;
      Event.xclient.format		= 32;
      Event.xclient.window		= Target;
      Event.xclient.data.l[0]		= DataType;
      Event.xclient.data.l[1]		= (long)event->xbutton.state;
      Event.xclient.data.l[2]		= Target;
      Event.xclient.data.l[3]		= 0;
      Event.xclient.data.l[4]		= 0;

      XSendEvent(dpy,DispatchWindow,True,NoEventMask,&Event);

      DataOK=0;
      return 1;
   }


/* FlDndInitialize
   * 
   * Must be called anywhere before the top level widget creation and the
   * main loop. Initialize global variables and bind the DndDispatch function
   * to the top level widget. Creates the cursors to be used in drag actions.
   * 
   * Last modification
   * 
   * Feb 01 1996 (v0.0) : created
   * Apr 01 1997 (v0.0) : modified to use with FORMS LIBRARY
*/
void   FlDndInitialize(FL_FORM *form)
   {
      int	 screen,i;
      Colormap colormap;
      Window	 root;
      XColor	 Black,White;

      dpy	= fl_get_display();
      screen	= DefaultScreen(dpy);
      colormap= DefaultColormap(dpy,screen);
      root	= DefaultRootWindow(dpy);


      Black.pixel=BlackPixel(dpy,screen);
      White.pixel=WhitePixel(dpy,screen);
      XQueryColor(dpy,colormap,&Black);
      XQueryColor(dpy,colormap,&White);

      for(i=1;i!=DndEND;i++)
      {
	 DndCursor[i].ImagePixmap=
	    XCreateBitmapFromData(dpy,root,
	    DndCursor[i].ImageData,
	    DndCursor[i].Width,
	    DndCursor[i].Height);
	 DndCursor[i].MaskPixmap=
	    XCreateBitmapFromData(dpy,root,
	    DndCursor[i].MaskData,
	    DndCursor[i].Width,
	    DndCursor[i].Height);
	 DndCursor[i].CursorID=
	    XCreatePixmapCursor(dpy,DndCursor[i].ImagePixmap,
	    DndCursor[i].MaskPixmap,
	    &Black,&White,
	    DndCursor[i].HotSpotX,
	    DndCursor[i].HotSpotY);
      }

      DndCursor[0].CursorID=XCreateFontCursor(dpy,XC_question_arrow);

      DndProtocol=XInternAtom(dpy,"DndProtocol",FALSE);
      DndSelection=XInternAtom(dpy,"DndSelection",FALSE);
      WM_STATE=XInternAtom(dpy,"WM_STATE",True);
      Dragging=0;
      DragPrecision=10;
      RootFlag=0;
   }


/* FlDndSetData
   * 
   * Updates the selection data.
   * 
   * Last modification
   * 
   * Feb 05 1996 (v0.0): created
   * Feb 29 1996 (v0.1): long sizes
   * Apr 01 1997 (v0.0) : modified to use with FORMS LIBRARY
*/
void FlDndSetData(int Type,unsigned char *Data,unsigned long Size)
{
   Window root		= DefaultRootWindow(dpy);
   int    AuxSize;

   if(DataOK) return;

/* Set the data type */
   DataType = Type>=DndEND ? 0 : Type ;

/* Set the data */
   AuxSize = Size > MAXINT ? MAXINT : (int)Size ;
   XChangeProperty(dpy,root,DndSelection,XA_STRING,8,
      PropModeReplace,Data,AuxSize);
   for(Size-=(unsigned long)AuxSize;Size;Size-=(unsigned long)AuxSize)
   {  Data+=AuxSize;
      AuxSize = Size > MAXINT ? MAXINT : (int)Size ;
      XChangeProperty(dpy,root,DndSelection,XA_STRING,8,
	 PropModeAppend,Data,AuxSize);
   }

/* Everything is now ok */
   DataOK=1;
}

/* FlDndGetData
   * 
   * Return a pointer to the current data. Se HOWTO for more details.
*/
void   FlDndGetData(unsigned char **Data,unsigned long *Size)
   {
      Display	*dpy		= fl_get_display();
      Window 	root		= DefaultRootWindow(dpy);

      Atom	ActualType;
      int	ActualFormat;
      unsigned long RemainingBytes;


      XGetWindowProperty(dpy,root,DndSelection,
	 0L,1000000L,
	 FALSE,AnyPropertyType,
	 &ActualType,&ActualFormat,
	 Size,&RemainingBytes,
	 Data);
   }

/* FlDndDataType 
   * 
   * Return information about the Dnd event received. If a non-dnd event is
   * passed, the function DndDataType returns DndNotDnd, and the others
   * return zero.
   * 
   * Last Updated:
   * 
   * Feb 02 1996 (v0.0) : created.
   * Apr 01 1997 (v0.0) : modified to use with FORMS LIBRARY
*/
int FlDndDataType(XEvent *event)
{
   int Type;

   if(event->xclient.message_type!=DndProtocol) return DndNotDnd;
   Type=(int)(event->xclient.data.l[0]);
   if(Type>=DndEND) Type=DndUnknown;
   return Type;
}


